@@ -231,6 +231,7 @@ external debug_hexdiff : buf -> buf -> string option -> int64 -> unit =
 external disconnect : bool -> unit = "ocaml_nbdkit_disconnect" [@@noalloc]
 external export_name : unit -> string = "ocaml_nbdkit_export_name"
 external is_tls : unit -> bool = "ocaml_nbdkit_is_tls"
+external name : unit -> string option = "ocaml_nbdkit_name"
 external nanosleep : int -> int -> unit = "ocaml_nbdkit_nanosleep"
 external parse_bool : string -> bool = "ocaml_nbdkit_parse_bool"
 external parse_delay : string -> string -> int * int =
@@ -252,4 +253,5 @@ external realpath : string -> string = "ocaml_nbdkit_realpath"
 external set_error : Unix.error -> unit = "ocaml_nbdkit_set_error" [@@noalloc]
 external shutdown : unit -> unit = "ocaml_nbdkit_shutdown" [@@noalloc]
 external stdio_safe : unit -> bool = "ocaml_nbdkit_stdio_safe"
+external timestamp : unit -> string = "ocaml_nbdkit_timestamp"
 external version : unit -> string = "ocaml_nbdkit_version"
